<?php
/**
 * WooCommerce Jilt
 *
 * This source file is subject to the GNU General Public License v3.0
 * that is bundled with this package in the file license.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.gnu.org/licenses/gpl-3.0.html
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@jilt.com so we can send you a copy immediately.
 *
 * DISCLAIMER
 *
 * Do not edit or add to this file if you wish to upgrade WooCommerce Jilt to newer
 * versions in the future. If you wish to customize WooCommerce Jilt for your
 * needs please refer to http://help.jilt.com/jilt-for-woocommerce
 *
 * @package   WC-Jilt/Checkout
 * @author    Jilt
 * @category  Frontend
 * @copyright Copyright (c) 2015-2017, SkyVerge, Inc.
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU General Public License v3.0
 */

defined( 'ABSPATH' ) or exit;

/**
 * Checkout Class
 *
 * Handles checkout page and orders that have been placed, but not yet paid for
 *
 * @since 1.0.0
 */
class WC_Jilt_Checkout_Handler {


	/**
	 * Setup class
	 *
	 * @since 1.0.0
	 */
	public function __construct() {

		$this->init();
	}


	/**
	 * Add hooks
	 *
	 * @since 1.0.0
	 */
	protected function init() {

		if ( SV_WC_Plugin_Compatibility::is_wc_version_gte_3_0() ) {

			add_filter( 'woocommerce_checkout_fields', array( $this, 'move_checkout_email_field' ), 1 );

		} else {

			// load customer data from session when a cart is recreated
			add_action( 'woocommerce_before_checkout_form', array( $this, 'load_customer_data_from_session' ) );
		}

		// maybe apply coupon code provided in the recovery URL
		add_action( 'woocommerce_cart_loaded_from_session', array( $this, 'maybe_apply_cart_recovery_coupon' ), 11 );

		// set order note content available and if pending recovery
		add_action( 'woocommerce_checkout_get_value', array( $this, 'maybe_set_order_note' ), 1, 2 );

		// handle placed orders
		add_action( 'woocommerce_checkout_order_processed', array( $this, 'checkout_order_processed' ) );

		// handle updating Jilt order data after a successful payment, for certain gateways
		add_filter( 'woocommerce_payment_successful_result', array( $this, 'maybe_update_jilt_order_on_successful_payment' ), 10, 2 );

		// handle marking an order as recovered when completing it from the pay page
		add_action( 'woocommerce_thankyou', array( $this, 'handle_pay_page_order_completion' ) );
	}


	/**
	 * Gets the cart checkout URL for Jilt
	 *
	 * Visiting this URL will load the associated cart from session/persistent cart
	 *
	 * @since 1.0.0
	 * @param string $jilt_cart_token Jilt cart token
	 * @param string $jilt_order_id Jilt order ID
	 * @return string
	 */
	public static function get_checkout_recovery_url( $jilt_cart_token, $jilt_order_id ) {

		$data = array(
			'order_id'   => $jilt_order_id,
			'cart_token' => $jilt_cart_token,
		);

		// encode
		$data = base64_encode( wp_json_encode( $data ) );

		// add hash for easier verification that the checkout URL hasn't been tampered with
		$hash = hash_hmac( 'sha256', $data, wc_jilt()->get_integration()->get_secret_key() );

		$url = self::get_jilt_wc_api_url();

		// returns URL like:
		// pretty permalinks enabled - https://example.com/wc-api/jilt?token=abc123&hash=xyz
		// pretty permalinks disabled - https://example.com?wc-api=jilt&token=abc123&hash=xyz
		return esc_url_raw( add_query_arg( array( 'token' => rawurlencode( $data ), 'hash' => $hash ), $url ) );
	}


	/**
	 * Maybe apply the recovery coupon provided in the recovery URL.
	 *
	 * @since 1.1.0
	 */
	public function maybe_apply_cart_recovery_coupon() {

		if ( WC_Jilt_Session::is_pending_recovery() && ! empty( $_REQUEST['coupon'] ) ) {

			$coupon_code = wc_clean( rawurldecode( $_REQUEST['coupon'] ) );

			if ( ! WC()->cart->has_discount( $coupon_code ) ) {
				WC()->cart->add_discount( $coupon_code );
			}
		}
	}


	/**
	 * Move the email field to the top of the checkout billing form.
	 *
	 * WC 3.0+ moved the email field to the bottom of the checkout form,
	 * which is less than ideal for capturing it. This method moves it
	 * to the top and makes it full-width.
	 *
	 * @since 1.1.0
	 * @param array $fields
	 * @return array
	 */
	public function move_checkout_email_field( $fields ) {

		if ( isset( $fields['billing'], $fields['billing']['billing_email'], $fields['billing']['billing_email']['priority'] ) ) {

			$email_field = $fields['billing']['billing_email'];
			unset( $fields['billing']['billing_email'] );

			$email_field['priority']  = 5;
			$email_field['class']     = array( 'form-row-wide' );
			$email_field['autofocus'] = true;

			$fields['billing'] = array_merge( array( 'billing_email' => $email_field ), $fields['billing'] );

			// adjust layout of postcode/phone fields
			if ( isset( $fields['billing']['billing_postcode'], $fields['billing']['billing_phone'] ) ) {

				// note this method is hooked in at priority 1, so other customizations *should* be safe to add additional classes since we've gone first
				$fields['billing']['billing_postcode']['class'] = array( 'form-row-first', 'address-field' );
				$fields['billing']['billing_phone']['class']    = array( 'form-row-last' );
			}

			// remove autofocus from billing first name (set to email above)
			if ( isset( $fields['billing']['billing_first_name'] ) && ! empty( $fields['billing']['billing_first_name']['autofocus'] ) ) {
				$fields['billing']['billing_first_name']['autofocus'] = false;
			}
		}

		return $fields;
	}


	/**
	 * When a customer visits the checkout recovery URL, load their data from
	 * the session and pre-fill the checkout form
	 *
	 * Note that WC 3.0+ handles this automatically when using setters in
	 * the WC_Customer class.
	 *
	 * TODO: Remove this (and default_checkout_value()) when WC 3.0+ is
	 * required. {MR 2017-03-28}
	 *
	 * @since 1.0.0
	 */
	public function load_customer_data_from_session() {

		if ( ! WC_Jilt_Session::is_pending_recovery() ) {
			return;
		}

		// add default value filter hooks for checkout fields
		$address_fields = array_merge( WC()->countries->get_address_fields(), WC()->countries->get_address_fields( '', 'shipping_' ) );

		foreach ( $address_fields as $field => $data ) {
			add_filter( 'default_checkout_' . $field, array( $this, 'default_checkout_value' ), 10, 2 );
		}
	}


	/**
	 * Get default checkout value from session
	 *
	 * @since 1.0.0
	 * @param mixed $value
	 * @param string $input
	 * @return mixed
	 */
	public function default_checkout_value( $value, $input ) {
		$input  = str_replace( 'billing_', '', $input );
		$method = "get_$input";
		if ( ! $value ) {
			// if there's a getter method available on WC_Customer for this field, use it
			if ( is_callable( array( WC()->customer, $method ) ) ) {
				$value = WC()->customer->{$method}();
			}
			// otherwise, fall back to session
			else {
				$customer = WC()->session->get('customer');
				if ( $customer && isset( $customer[ $input ] ) ) {
					$value = $customer[ $input ];
				}
			}
		}
		return $value;
	}


	/**
	 * Maybe set the order note when there is a pending recovery
	 * with a previous value for the order note from the Jilt order.
	 *
	 * Note that unlike customer data (email, etc). this will not persist
	 * when/if the customer navigates away from the checkout page. We want to avoid
	 * a situation where the customer feels like they can't change the value of
	 * the order note field after it's been populated for them.
	 *
	 * @since 1.1.0
	 * @param string $value null
	 * @param string $input input field name
	 * @return null|string
	 */
	public function maybe_set_order_note( $value, $input ) {

		// target order comments input when a pending recovery order has an order note present
		if ( 'order_comments' !== $input || ! WC_Jilt_Session::is_pending_recovery() || ! $order_note = WC()->session->get( 'wc_jilt_order_note' ) ) {
			return $value;
		}

		unset( WC()->session->wc_jilt_order_note );

		return $order_note;
	}


	/**
	 * This is called once the checkout has been processed and an order has been created.
	 * Does not necessarily mean that the order has been paid for.
	 *
	 * @since 1.0.0
	 * @param int $order_id order ID
	 */
	public function checkout_order_processed( $order_id ) {

		if ( wc_jilt()->get_integration()->is_disabled() ) {
			return;
		}

		$cart_token    = WC_Jilt_Session::get_cart_token();
		$jilt_order_id = WC_Jilt_Session::get_jilt_order_id();

		// bail out if this cart is not associated with a Jilt order
		if ( ! $jilt_order_id ) {
			return;
		}

		// save Jilt order ID and cart token to order meta
		update_post_meta( $order_id, '_wc_jilt_cart_token', $cart_token );
		update_post_meta( $order_id, '_wc_jilt_order_id', $jilt_order_id );

		// mark as recovered
		if ( WC_Jilt_Session::is_pending_recovery() ) {

			$this->mark_order_as_recovered( $order_id );
		}

		// update Jilt order details
		try {

			$order = new WC_Jilt_Order( $order_id );
			$this->get_api()->update_order( $jilt_order_id, $this->get_jilt_order_data( $order ) );

		} catch ( SV_WC_API_Exception $exception ) {

			wc_jilt()->get_logger()->error( "Error communicating with Jilt: {$exception->getMessage()}" );
		}

		// remove Jilt order ID from session and user meta
		WC_Jilt_Session::unset_jilt_order_data();
	}


	/**
	 * Handle updating the Jilt order when order details aren't available until
	 * *after* payment is received. Gateways like Amazon Payments Advanced and
	 * other on-site/iframed gateways act this way and otherwise result in a lot
	 * of placed orders with empty order data in Jilt.
	 *
	 * Important: this is a non-standard use of a filter used in an action context,
	 * but I preferred to use this over the woocommerce_thankyou action since that
	 * requires it to be present on the template. {MR 2016-12-06}
	 *
	 * @since 1.0.7
	 * @param array $result payment successful result
	 * @param int $order_id
	 * @return array
	 */
	public function maybe_update_jilt_order_on_successful_payment( $result, $order_id ) {

		if ( wc_jilt()->get_integration()->is_disabled() ) {
			return $result;
		}

		if ( ! $jilt_order_id = get_post_meta( $order_id, '_wc_jilt_order_id', true ) ) {
			return $result;
		}

		try {

			$order = new WC_Jilt_Order( $order_id );
			$this->get_api()->update_order( $jilt_order_id, $this->get_jilt_order_data( $order ) );

		} catch ( SV_WC_API_Exception $exception ) {

			wc_jilt()->get_logger()->error( "Error communicating with Jilt: {$exception->getMessage()}" );
		}

		return $result;
	}


	/**
	 * When a customer completes a pending recovery from the pay page (e.g an order
	 * originally placed with an off-site gateway then later completed via an
	 * on-site gateway), mark it as recovered.
	 *
	 * @since 1.0.7
	 * @param $order_id
	 */
	public function handle_pay_page_order_completion( $order_id ) {

		if ( WC_Jilt_Session::is_pending_recovery() ) {

			$this->mark_order_as_recovered( $order_id );

			WC_Jilt_Session::unset_jilt_order_data();
		}
	}


	/**
	 * Mark an order as recovered by Jilt
	 *
	 * @since 1.0.1
	 * @param int|string $order_id order ID
	 */
	protected function mark_order_as_recovered( $order_id ) {

		$order = wc_get_order( $order_id );

		if ( ! $order instanceof WC_Order ) {
			return;
		}

		update_post_meta( SV_WC_Order_Compatibility::get_prop( $order, 'id' ), '_wc_jilt_recovered', true );

		$order->add_order_note( __( 'Order recovered by Jilt.', 'jilt-for-woocommerce' ) );
	}


	/**
	 * Get the order data
	 *
	 * @since 1.2.0
	 * @return array associative array of order data formatted for Jilt
	 */
	private function get_jilt_order_data( $order ) {
		$data = $order->get_jilt_order_data();
		$data['client_session']  = WC_Jilt_Session::get_client_session();

		return $data;
	}


	/**
	 * Helper method to improve the readability of methods calling the API
	 *
	 * @since 1.0.0
	 * @return \WC_Jilt_API instance
	 */
	private function get_api() {
		return wc_jilt()->get_integration()->get_api();
	}


	/**
	 * Return the WC API URL for handling Jilt recovery links by accounting
	 * for whether pretty permalinks are enabled or not.
	 *
	 * @since 1.1.0
	 * @return string
	 */
	private static function get_jilt_wc_api_url() {

		$scheme = wc_site_is_https() ? 'https' : 'http';

		return get_option( 'permalink_structure' )
			? get_home_url( null, 'wc-api/jilt', $scheme )
			: add_query_arg( 'wc-api', 'jilt', get_home_url( null, null, $scheme ) );
	}


}
